package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.ConstructorDeclaration;
import org.eclipse.jdt.internal.compiler.ast.LambdaExpression;
import org.eclipse.jdt.internal.compiler.ast.QualifiedNameReference;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.codegen.ConstantPool;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemFieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticArgumentBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class MethodScope extends BlockScope {
  public ReferenceContext referenceContext;
  
  public boolean isStatic;
  
  public boolean isConstructorCall = false;
  
  public FieldBinding initializedField;
  
  public int lastVisibleFieldID = -1;
  
  public int analysisIndex;
  
  public boolean isPropagatingInnerClassEmulation;
  
  public int lastIndex = 0;
  
  public long[] definiteInits = new long[4];
  
  public long[][] extraDefiniteInits = new long[4][];
  
  public SyntheticArgumentBinding[] extraSyntheticArguments;
  
  public boolean hasMissingSwitchDefault;
  
  public MethodScope(Scope parent, ReferenceContext context, boolean isStatic) {
    super(2, parent);
    this.locals = new LocalVariableBinding[5];
    this.referenceContext = context;
    this.isStatic = isStatic;
    this.startIndex = 0;
  }
  
  public MethodScope(Scope parent, ReferenceContext context, boolean isStatic, int lastVisibleFieldID) {
    this(parent, context, isStatic);
    this.lastVisibleFieldID = lastVisibleFieldID;
  }
  
  String basicToString(int tab) {
    String newLine = "\n";
    for (int i = tab; --i >= 0;)
      newLine = String.valueOf(newLine) + "\t"; 
    String s = String.valueOf(newLine) + "--- Method Scope ---";
    newLine = String.valueOf(newLine) + "\t";
    s = String.valueOf(s) + newLine + "locals:";
    for (int j = 0; j < this.localIndex; j++)
      s = String.valueOf(s) + newLine + "\t" + this.locals[j].toString(); 
    s = String.valueOf(s) + newLine + "startIndex = " + this.startIndex;
    s = String.valueOf(s) + newLine + "isConstructorCall = " + this.isConstructorCall;
    s = String.valueOf(s) + newLine + "initializedField = " + this.initializedField;
    s = String.valueOf(s) + newLine + "lastVisibleFieldID = " + this.lastVisibleFieldID;
    s = String.valueOf(s) + newLine + "referenceContext = " + this.referenceContext;
    return s;
  }
  
  private void checkAndSetModifiersForConstructor(MethodBinding methodBinding) {
    int modifiers = methodBinding.modifiers;
    ReferenceBinding declaringClass = methodBinding.declaringClass;
    if ((modifiers & 0x400000) != 0)
      problemReporter().duplicateModifierForMethod(declaringClass, (AbstractMethodDeclaration)this.referenceContext); 
    if ((((ConstructorDeclaration)this.referenceContext).bits & 0x80) != 0) {
      int flags;
      if ((flags = declaringClass.modifiers & 0x4005) != 0)
        if ((flags & 0x4000) != 0) {
          modifiers &= 0xFFFFFFF8;
          modifiers |= 0x2;
        } else {
          modifiers &= 0xFFFFFFF8;
          modifiers |= flags;
        }  
    } 
    int realModifiers = modifiers & 0xFFFF;
    if (declaringClass.isEnum() && (((ConstructorDeclaration)this.referenceContext).bits & 0x80) == 0) {
      if ((realModifiers & 0xFFFFF7FD) != 0) {
        problemReporter().illegalModifierForEnumConstructor((AbstractMethodDeclaration)this.referenceContext);
        modifiers &= 0xFFFF0802;
      } else if ((((AbstractMethodDeclaration)this.referenceContext).modifiers & 0x800) != 0) {
        problemReporter().illegalModifierForMethod((AbstractMethodDeclaration)this.referenceContext);
      } 
      modifiers |= 0x2;
    } else if ((realModifiers & 0xFFFFF7F8) != 0) {
      problemReporter().illegalModifierForMethod((AbstractMethodDeclaration)this.referenceContext);
      modifiers &= 0xFFFF0807;
    } else if ((((AbstractMethodDeclaration)this.referenceContext).modifiers & 0x800) != 0) {
      problemReporter().illegalModifierForMethod((AbstractMethodDeclaration)this.referenceContext);
    } 
    int accessorBits = realModifiers & 0x7;
    if ((accessorBits & accessorBits - 1) != 0) {
      problemReporter().illegalVisibilityModifierCombinationForMethod(declaringClass, (AbstractMethodDeclaration)this.referenceContext);
      if ((accessorBits & 0x1) != 0) {
        if ((accessorBits & 0x4) != 0)
          modifiers &= 0xFFFFFFFB; 
        if ((accessorBits & 0x2) != 0)
          modifiers &= 0xFFFFFFFD; 
      } else if ((accessorBits & 0x4) != 0 && (accessorBits & 0x2) != 0) {
        modifiers &= 0xFFFFFFFD;
      } 
    } 
    methodBinding.modifiers = modifiers;
  }
  
  private void checkAndSetModifiersForMethod(MethodBinding methodBinding) {
    int modifiers = methodBinding.modifiers;
    ReferenceBinding declaringClass = methodBinding.declaringClass;
    if ((modifiers & 0x400000) != 0)
      problemReporter().duplicateModifierForMethod(declaringClass, (AbstractMethodDeclaration)this.referenceContext); 
    int realModifiers = modifiers & 0xFFFF;
    if (declaringClass.isInterface()) {
      int expectedModifiers = 1025;
      boolean isDefaultMethod = ((modifiers & 0x10000) != 0);
      boolean reportIllegalModifierCombination = false;
      boolean isJDK18orGreater = false;
      if ((compilerOptions()).sourceLevel >= 3407872L && !declaringClass.isAnnotationType()) {
        expectedModifiers |= 0x10808;
        isJDK18orGreater = true;
        if (!methodBinding.isAbstract()) {
          reportIllegalModifierCombination = (isDefaultMethod && methodBinding.isStatic());
        } else {
          reportIllegalModifierCombination = !(!isDefaultMethod && !methodBinding.isStatic());
          if (methodBinding.isStrictfp())
            problemReporter().illegalAbstractModifierCombinationForMethod((AbstractMethodDeclaration)this.referenceContext); 
        } 
        if (reportIllegalModifierCombination)
          problemReporter().illegalModifierCombinationForInterfaceMethod((AbstractMethodDeclaration)this.referenceContext); 
        if (isDefaultMethod)
          realModifiers |= 0x10000; 
      } 
      if ((realModifiers & (expectedModifiers ^ 0xFFFFFFFF)) != 0)
        if ((declaringClass.modifiers & 0x2000) != 0) {
          problemReporter().illegalModifierForAnnotationMember((AbstractMethodDeclaration)this.referenceContext);
        } else {
          problemReporter().illegalModifierForInterfaceMethod((AbstractMethodDeclaration)this.referenceContext, isJDK18orGreater);
        }  
      return;
    } 
    if ((realModifiers & 0xFFFFF2C0) != 0) {
      problemReporter().illegalModifierForMethod((AbstractMethodDeclaration)this.referenceContext);
      modifiers &= 0xFFFF0D3F;
    } 
    int accessorBits = realModifiers & 0x7;
    if ((accessorBits & accessorBits - 1) != 0) {
      problemReporter().illegalVisibilityModifierCombinationForMethod(declaringClass, (AbstractMethodDeclaration)this.referenceContext);
      if ((accessorBits & 0x1) != 0) {
        if ((accessorBits & 0x4) != 0)
          modifiers &= 0xFFFFFFFB; 
        if ((accessorBits & 0x2) != 0)
          modifiers &= 0xFFFFFFFD; 
      } else if ((accessorBits & 0x4) != 0 && (accessorBits & 0x2) != 0) {
        modifiers &= 0xFFFFFFFD;
      } 
    } 
    if ((modifiers & 0x400) != 0) {
      int incompatibleWithAbstract = 2362;
      if ((modifiers & incompatibleWithAbstract) != 0)
        problemReporter().illegalAbstractModifierCombinationForMethod(declaringClass, (AbstractMethodDeclaration)this.referenceContext); 
      if (!methodBinding.declaringClass.isAbstract())
        problemReporter().abstractMethodInAbstractClass((SourceTypeBinding)declaringClass, (AbstractMethodDeclaration)this.referenceContext); 
    } 
    if ((modifiers & 0x100) != 0 && (modifiers & 0x800) != 0)
      problemReporter().nativeMethodsCannotBeStrictfp(declaringClass, (AbstractMethodDeclaration)this.referenceContext); 
    if ((realModifiers & 0x8) != 0 && declaringClass.isNestedType() && !declaringClass.isStatic())
      problemReporter().unexpectedStaticModifierForMethod(declaringClass, (AbstractMethodDeclaration)this.referenceContext); 
    methodBinding.modifiers = modifiers;
  }
  
  public void checkUnusedParameters(MethodBinding method) {
    if (method.isAbstract() || (
      method.isImplementing() && !(compilerOptions()).reportUnusedParameterWhenImplementingAbstract) || (
      method.isOverriding() && !method.isImplementing() && !(compilerOptions()).reportUnusedParameterWhenOverridingConcrete) || 
      method.isMain())
      return; 
    for (int i = 0, maxLocals = this.localIndex; i < maxLocals; i++) {
      LocalVariableBinding local = this.locals[i];
      if (local == null || (local.tagBits & 0x400L) == 0L)
        break; 
      if (local.useFlag == 0)
        if ((local.declaration.bits & 0x40000000) != 0)
          problemReporter().unusedArgument(local.declaration);  
    } 
  }
  
  public void computeLocalVariablePositions(int initOffset, CodeStream codeStream) {
    this.offset = initOffset;
    this.maxOffset = initOffset;
    int ilocal = 0, maxLocals = this.localIndex;
    while (ilocal < maxLocals) {
      LocalVariableBinding local = this.locals[ilocal];
      if (local == null || (local.tagBits & 0x400L) == 0L)
        break; 
      codeStream.record(local);
      local.resolvedPosition = this.offset;
      if (TypeBinding.equalsEquals(local.type, (TypeBinding)TypeBinding.LONG) || TypeBinding.equalsEquals(local.type, (TypeBinding)TypeBinding.DOUBLE)) {
        this.offset += 2;
      } else {
        this.offset++;
      } 
      if (this.offset > 255)
        problemReporter().noMoreAvailableSpaceForArgument(local, (ASTNode)local.declaration); 
      ilocal++;
    } 
    if (this.extraSyntheticArguments != null)
      for (int iarg = 0, maxArguments = this.extraSyntheticArguments.length; iarg < maxArguments; iarg++) {
        SyntheticArgumentBinding argument = this.extraSyntheticArguments[iarg];
        argument.resolvedPosition = this.offset;
        if (TypeBinding.equalsEquals(argument.type, (TypeBinding)TypeBinding.LONG) || TypeBinding.equalsEquals(argument.type, (TypeBinding)TypeBinding.DOUBLE)) {
          this.offset += 2;
        } else {
          this.offset++;
        } 
        if (this.offset > 255)
          problemReporter().noMoreAvailableSpaceForArgument((LocalVariableBinding)argument, (ASTNode)this.referenceContext); 
      }  
    computeLocalVariablePositions(ilocal, this.offset, codeStream);
  }
  
  MethodBinding createMethod(AbstractMethodDeclaration method) {
    this.referenceContext = (ReferenceContext)method;
    method.scope = this;
    SourceTypeBinding declaringClass = (referenceType()).binding;
    int modifiers = method.modifiers | 0x2000000;
    if (method.isConstructor()) {
      if (method.isDefaultConstructor())
        modifiers |= 0x4000000; 
      method.binding = new MethodBinding(modifiers, null, null, (ReferenceBinding)declaringClass);
      checkAndSetModifiersForConstructor(method.binding);
    } else {
      if (declaringClass.isInterface())
        if (method.isDefaultMethod() || method.isStatic()) {
          modifiers |= 0x1;
        } else {
          modifiers |= 0x401;
        }  
      method.binding = 
        new MethodBinding(modifiers, method.selector, null, null, null, (ReferenceBinding)declaringClass);
      checkAndSetModifiersForMethod(method.binding);
    } 
    this.isStatic = method.binding.isStatic();
    Argument[] argTypes = method.arguments;
    int argLength = (argTypes == null) ? 0 : argTypes.length;
    long sourceLevel = (compilerOptions()).sourceLevel;
    if (argLength > 0) {
      Argument argument = argTypes[--argLength];
      if (argument.isVarArgs() && sourceLevel >= 3211264L)
        method.binding.modifiers |= 0x80; 
      if (CharOperation.equals(argument.name, ConstantPool.This))
        problemReporter().illegalThisDeclaration(argument); 
      while (--argLength >= 0) {
        argument = argTypes[argLength];
        if (argument.isVarArgs() && sourceLevel >= 3211264L)
          problemReporter().illegalVararg(argument, method); 
        if (CharOperation.equals(argument.name, ConstantPool.This))
          problemReporter().illegalThisDeclaration(argument); 
      } 
    } 
    if (method.receiver != null) {
      if (sourceLevel <= 3342336L)
        problemReporter().illegalSourceLevelForThis((Argument)method.receiver); 
      if (method.receiver.annotations != null)
        method.bits |= 0x100000; 
    } 
    TypeParameter[] typeParameters = method.typeParameters();
    if (typeParameters == null || typeParameters.length == 0) {
      method.binding.typeVariables = Binding.NO_TYPE_VARIABLES;
    } else {
      method.binding.typeVariables = createTypeVariables(typeParameters, (Binding)method.binding);
      method.binding.modifiers |= 0x40000000;
    } 
    return method.binding;
  }
  
  public FieldBinding findField(TypeBinding receiverType, char[] fieldName, InvocationSite invocationSite, boolean needResolve) {
    FieldBinding field = super.findField(receiverType, fieldName, invocationSite, needResolve);
    if (field == null)
      return null; 
    if (!field.isValidBinding())
      return field; 
    if (receiverType.isInterface() && invocationSite.isQualifiedSuper())
      return (FieldBinding)new ProblemFieldBinding(
          field, 
          field.declaringClass, 
          fieldName, 
          28); 
    if (field.isStatic())
      return field; 
    if (!this.isConstructorCall || TypeBinding.notEquals(receiverType, (TypeBinding)enclosingSourceType()))
      return field; 
    if (invocationSite instanceof org.eclipse.jdt.internal.compiler.ast.SingleNameReference)
      return (FieldBinding)new ProblemFieldBinding(
          field, 
          field.declaringClass, 
          fieldName, 
          6); 
    if (invocationSite instanceof QualifiedNameReference) {
      QualifiedNameReference name = (QualifiedNameReference)invocationSite;
      if (name.binding == null)
        return (FieldBinding)new ProblemFieldBinding(
            field, 
            field.declaringClass, 
            fieldName, 
            6); 
    } 
    return field;
  }
  
  public boolean isInsideConstructor() {
    return this.referenceContext instanceof ConstructorDeclaration;
  }
  
  public boolean isInsideInitializer() {
    return this.referenceContext instanceof TypeDeclaration;
  }
  
  public boolean isLambdaScope() {
    return this.referenceContext instanceof LambdaExpression;
  }
  
  public boolean isInsideInitializerOrConstructor() {
    return !(!(this.referenceContext instanceof TypeDeclaration) && 
      !(this.referenceContext instanceof ConstructorDeclaration));
  }
  
  public ProblemReporter problemReporter() {
    ProblemReporter problemReporter = (referenceCompilationUnit()).problemReporter;
    problemReporter.referenceContext = this.referenceContext;
    return problemReporter;
  }
  
  public final int recordInitializationStates(FlowInfo flowInfo) {
    if ((flowInfo.tagBits & 0x1) != 0)
      return -1; 
    UnconditionalFlowInfo unconditionalFlowInfo = flowInfo.unconditionalInitsWithoutSideEffect();
    long[] extraInits = (unconditionalFlowInfo.extra == null) ? 
      null : unconditionalFlowInfo.extra[0];
    long inits = unconditionalFlowInfo.definiteInits;
    for (int i = this.lastIndex; --i >= 0;) {
      if (this.definiteInits[i] == inits) {
        long[] otherInits = this.extraDefiniteInits[i];
        if (extraInits != null && otherInits != null) {
          if (extraInits.length == otherInits.length) {
            int j = 0, max = extraInits.length;
            while (true) {
              if (j >= max)
                return i; 
              if (extraInits[j] != otherInits[j])
                break; 
              j++;
            } 
          } 
          continue;
        } 
        if (extraInits == null && otherInits == null)
          return i; 
      } 
    } 
    if (this.definiteInits.length == this.lastIndex) {
      System.arraycopy(
          this.definiteInits, 
          0, 
          this.definiteInits = new long[this.lastIndex + 20], 
          0, 
          this.lastIndex);
      System.arraycopy(
          this.extraDefiniteInits, 
          0, 
          this.extraDefiniteInits = new long[this.lastIndex + 20][], 
          0, 
          this.lastIndex);
    } 
    this.definiteInits[this.lastIndex] = inits;
    if (extraInits != null) {
      this.extraDefiniteInits[this.lastIndex] = new long[extraInits.length];
      System.arraycopy(
          extraInits, 
          0, 
          this.extraDefiniteInits[this.lastIndex], 
          0, 
          extraInits.length);
    } 
    return this.lastIndex++;
  }
  
  public AbstractMethodDeclaration referenceMethod() {
    if (this.referenceContext instanceof AbstractMethodDeclaration)
      return (AbstractMethodDeclaration)this.referenceContext; 
    return null;
  }
  
  public MethodBinding referenceMethodBinding() {
    if (this.referenceContext instanceof LambdaExpression)
      return ((LambdaExpression)this.referenceContext).binding; 
    if (this.referenceContext instanceof AbstractMethodDeclaration)
      return ((AbstractMethodDeclaration)this.referenceContext).binding; 
    return null;
  }
  
  public TypeDeclaration referenceType() {
    ClassScope scope = enclosingClassScope();
    return (scope == null) ? null : scope.referenceContext;
  }
  
  void resolveTypeParameter(TypeParameter typeParameter) {
    typeParameter.resolve(this);
  }
  
  public boolean hasDefaultNullnessFor(int location) {
    if (this.referenceContext instanceof AbstractMethodDeclaration) {
      MethodBinding binding = ((AbstractMethodDeclaration)this.referenceContext).binding;
      if (binding != null && binding.defaultNullness != 0)
        return ((binding.defaultNullness & location) != 0); 
    } 
    return this.parent.hasDefaultNullnessFor(location);
  }
}
